/*
 * Copyright (C) 2006,2008 Johan Maasing johan at zoom.nu Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package nu.zoom.catonine.swing.tail;

import java.awt.Font;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;

import javax.swing.text.Style;
import javax.swing.text.StyleConstants;

import nu.zoom.catonine.stylerule.StyleRule;
import nu.zoom.catonine.stylerule.StyleRules;

/**
 * Styled document that listens for new lines to add and colors the lines
 * according to the selected style rules.
 * 
 * @author "Johan Maasing" &lt;johan@zoom.nu&gt;
 */
public abstract class StyleRulesDocument extends ParagraphStyleDocument {

	private static final long serialVersionUID = 1L;
	final static String DEFAULT_CHARACTER_STYLE_NAME = "defaultCharacterStyleName";
	private StyleRules configuration = null;

	public StyleRulesDocument() {
		super();
		addStyle(DEFAULT_CHARACTER_STYLE_NAME, getStyle("default"));
	}

	protected void setConfiguration(StyleRules configuration) {
		this.configuration = configuration;
		setupStyles();
	}

	/**
	 * Get the styled document style name that should be used for the given log
	 * block. Matches the given string against the style rules to determine
	 * which style to use.
	 * 
	 * @param logblock
	 *            The string that should be styled.
	 * @return The style. Null if the block should be supressed, i.e. not drawn
	 *         at all.
	 */
	protected String getStyleName(String logblock) {
		StyleRule match = null;
		if (configuration != null) {
			Iterator<StyleRule> patternIterator = configuration.getPatternStyles().iterator();
			while ((match == null) && patternIterator.hasNext()) {
				StyleRule config = patternIterator.next();
				if (!config.isDisabled()) {
					Matcher matcher = config.getPattern().matcher(logblock);
					if (matcher.matches()) {
						match = config;
					}
				}
			}
		}
		if (match != null) {
			if (!match.isSupress()) {
				return match.getGuid();
			} else {
				return null;
			}
		} else {
			return DEFAULT_CHARACTER_STYLE_NAME;
		}
	}

	@Override
	protected synchronized void setupStyles() {
		Enumeration<?> styleNameEnumeration = getStyleNames();
		while (styleNameEnumeration.hasMoreElements()) {
			final String styleName = (String) styleNameEnumeration.nextElement();
			removeStyle(styleName);
		}
		super.setupStyles();
		final Style defaultStyle = addStyle(DEFAULT_CHARACTER_STYLE_NAME, null);
		StyleConstants.setFontFamily(defaultStyle, Font.MONOSPACED);
		
		if (configuration != null) {
			List<StyleRule> configurations = configuration.getPatternStyles();
			if (configurations != null) {
				for (StyleRule styleRule : configurations) {
					Style configStyle = addStyle(styleRule.getGuid(), defaultStyle);
					if (styleRule.getBackground() != null) {
						StyleConstants.setBackground(configStyle, styleRule.getBackground());
					}
					if (styleRule.getForeground() != null) {
						StyleConstants.setForeground(configStyle, styleRule.getForeground());
					}
				}
			}
		}
	}
}
