/*
Copyright 2010 Johan Maasing

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package nu.zoom.catonine.swing.tail;

import java.util.LinkedList;
import java.util.List;

import javax.swing.JSeparator;
import javax.swing.text.BadLocationException;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;

import nu.zoom.catonine.tail.TailerListener.LogEntry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Document that remembers document positions in relation to file positions.
 * 
 * @author "Johan Maasing" &lt;johan@zoom.nu&gt;
 * 
 */
public class TailingDocument extends StyleRulesDocument implements TimestampDocument {
	private final static char[] EOL_ARRAY = "\n".toCharArray();

	private static final long serialVersionUID = 1L;
	private static final String STAMP_STYLE_NAME = "nu.zoom.catonine.swing.tail.STAMP_STYLE_NAME";
	private final Log log = LogFactory.getLog(getClass());
	// Content MUST always be linked with lowest file position first (head) and
	// ascending.
	LinkedList<DocumentLogEntry> content = new LinkedList<TailingDocument.DocumentLogEntry>();

	public void lines(final List<LogEntry> lines) {
		log.debug("Adding " + lines.size() + " entries");
		final LinkedList<LogEntry> coalescedLines = coalesceLines(lines);
		log.debug("Coalesced lines to " + coalescedLines.size() + " entries");

		// Check if we need to update the last displayed entry with a new entry
		// from the tailer
		final LogEntry firstEntry = coalescedLines.peekFirst();
		removeUpdatedEntries(firstEntry);

		// We can now batch add all the coalesced entries
		final LinkedList<ElementSpec> batch = new LinkedList<ElementSpec>();
		int offsetCounter = getLength();
		for (final LogEntry logEntry : coalescedLines) {
			offsetCounter += batchEntry(offsetCounter, batch, logEntry);
		}
		ElementSpec[] batchArray = batch.toArray(new ElementSpec[batch.size()]);
		try {
			log.debug("Inserting batch of " + batchArray.length + " element specifications");
			insert(getLength(), batchArray);
		} catch (BadLocationException e) {
			log.error("Unable to insert batch", e);
		}
	}

	/**
	 * @param logEntry
	 */
	private int batchEntry(final int offsetCounter, final LinkedList<ElementSpec> batch, final LogEntry logEntry) {

		final String entryContent = logEntry.getContent();
		final String styleName = getStyleName(entryContent);
		int offsetChange = 0;

		// Style name == null means supress this entry
		if (styleName != null) {
			final Style style = getStyle(styleName);
			if (style != null) {
				final boolean endsWithNewLine = entryContent.endsWith("\n");
				String[] split = entryContent.split("\\n");
				if (split != null) {
					int entryLength = 0;
					for (int i = 0; i < split.length; i++) {
						// TODO: Replace splitting with regexp find for \n and
						// then construct element specs with offsets and length
						// into the chararray.
						// Splitting removes newlines, the last (or only) part
						// might or might not have ended in a newline. This also
						// removes consequtive newlines inside the logentry
						char[] charArray =
								((endsWithNewLine || (i < split.length - 1)) ? (split[i] + "\n") : split[i])
										.toCharArray();
						addLineBreak(batch);
						batch.add(new ElementSpec(style, ElementSpec.ContentType, charArray, 0, charArray.length));
						entryLength += charArray.length;
					}
					DocumentLogEntry documentLogEntry = new DocumentLogEntry();
					documentLogEntry.entryFileStart = logEntry.getPosition();
					documentLogEntry.styleGUID = styleName;
					documentLogEntry.offset = offsetCounter;
					documentLogEntry.length = entryLength;
					offsetChange = documentLogEntry.length;
					this.content.add(documentLogEntry);
				}
			}
		}
		return offsetChange;
	}

	private LinkedList<LogEntry> coalesceLines(List<LogEntry> list) {
		final LinkedList<LogEntry> result = new LinkedList<LogEntry>();
		for (final LogEntry newEntry : list) {
			final LogEntry existingEntry = result.peekLast();
			if (existingEntry != null) {
				final long existingPosition = existingEntry.getPosition();
				final long newPosition = newEntry.getPosition();
				if (existingPosition == newPosition) {
					result.removeLast();
				} else if (existingPosition > newPosition) {
					log.warn("Coalesce lines detected out of order entries");
				}
			}
			result.add(newEntry);
		}
		return result;
	}

	/**
	 * Checks if the log entry indicates a file position we already have
	 * displayed in the document.
	 */
	private synchronized void removeUpdatedEntries(final LogEntry logEntry) {
		DocumentLogEntry lastEntry = this.content.peekLast();
		boolean keepGoing = lastEntry != null;
		while (keepGoing) {
			if (lastEntry.entryFileStart >= logEntry.getPosition()) {
				DocumentLogEntry removedEntry = this.content.removeLast();
				try {
					remove(removedEntry.offset, removedEntry.length);
				} catch (BadLocationException e) {
					log.error(e);
				}
				lastEntry = this.content.peekLast();
			} else {
				keepGoing = false;
			}
		}
	}

	/**
	 * Indirect callback from the tailer.
	 */
	public void tailerHasBeenReset() {
		// Clear the document
		content.clear();
		int start = 0;
		int length = getLength();
		try {
			remove(start, length);
		} catch (BadLocationException e) {
			// if this happens there is something seriously wrong
			// with the lock on the document
			throw new IllegalStateException(e);
		}
	}

	@Override
	public void addStampAfterLast() {
		DocumentLogEntry documentLogEntry = new DocumentLogEntry();
		documentLogEntry.synthetic = true;
		documentLogEntry.styleGUID = STAMP_STYLE_NAME;
		documentLogEntry.offset = getLength();
		documentLogEntry.length = EOL_ARRAY.length;
		SimpleAttributeSet stampStyle = new SimpleAttributeSet(getStyle(DEFAULT_CHARACTER_STYLE_NAME));
		StyleConstants.setComponent(stampStyle, new JSeparator(JSeparator.HORIZONTAL));
		final ElementSpec contentElementSpec =
				new ElementSpec(stampStyle, ElementSpec.ContentType, EOL_ARRAY, 0, EOL_ARRAY.length);

		LinkedList<ElementSpec> batch = new LinkedList<ElementSpec>();
		batch.add(linebreakParagraphEnd);
		batch.add(linebreakParagraphStart);
		batch.add(contentElementSpec);
		try {
			insert(documentLogEntry.offset, batch.toArray(new ElementSpec[batch.size()]));
		} catch (BadLocationException e) {
			log.error("Unable to append entry - concurrency error?", e);
		}
		this.content.addLast(documentLogEntry);
	}

	static class DocumentLogEntry {
		boolean synthetic = false;
		int offset;
		int length;
		long entryFileStart;
		String styleGUID;

		@Override
		public String toString() {
			return "DocumentLogEntry [offset=" + offset + ", length=" + length + ", entryFileStart=" + entryFileStart
					+ ", styleGUID=" + styleGUID + "]";
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + (int) (entryFileStart ^ (entryFileStart >>> 32));
			result = prime * result + ((styleGUID == null) ? 0 : styleGUID.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			DocumentLogEntry other = (DocumentLogEntry) obj;
			if (entryFileStart != other.entryFileStart)
				return false;
			if (styleGUID == null) {
				if (other.styleGUID != null)
					return false;
			} else if (!styleGUID.equals(other.styleGUID))
				return false;
			return true;
		}

	}

}
